(*====================================================================
    File crt_win32.inc

    Components library for Delphi, Kylix and BCB++.
    Procedures and functions.

    Copyright  2002-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: crt_win32.inc,v 1.3 2008/03/04 18:51:39 ariafan Exp $
======================================================================*)

{*******************************************************}
{       Turbo Pascal Runtime Library                    }
{       CRT Interface Unit                              }
{       Copyright (C) 1988,92 Borland International     }
{*******************************************************}

{$B-}

interface

uses
  Windows;

const

{ CRT modes }
  BW40          = 0;            { 40x25 B/W on Color Adapter }
  CO40          = 1;            { 40x25 Color on Color Adapter }
  BW80          = 2;            { 80x25 B/W on Color Adapter }
  CO80          = 3;            { 80x25 Color on Color Adapter }
  Mono          = 7;            { 80x25 on Monochrome Adapter }
  Font8x8       = 256;          { Add-in for ROM font }
{ Mode constants for 3.0 compatibility }
  C40           = CO40;
  C80           = CO80;

{ Foreground and background color constants }
  Black         = 0;
  Blue          = 1;
  Green         = 2;
  Cyan          = 3;
  Red           = 4;
  Magenta       = 5;
  Brown         = 6;
  LightGray     = 7;
{ Foreground color constants }
  DarkGray      = 8;
  LightBlue     = 9;
  LightGreen    = 10;
  LightCyan     = 11;
  LightRed      = 12;
  LightMagenta  = 13;
  Yellow        = 14;
  White         = 15;
{ Add-in for blinking }
  Blink         = 128;

var

{ Interface variables }
  {$IFDEF fpas}
  CheckBreak: Boolean;    { Enable Ctrl-Break }
  CheckEOF: Boolean;      { Enable Ctrl-Z }
  DirectVideo: Boolean;   { Enable direct video addressing }
  CheckSnow: Boolean;     { Enable snow filtering }
  LastMode: Word;         { Current text mode }
  {$ENDIF}
  _TextAttr: Byte;        { Current text attribute }
  //WindMin: Word;        { Window upper left coordinates }
  //WindMax: Word;        { Window lower right coordinates }

//=============================================================================
// Unsupported variables emulation
function TextAttr: Byte;
function WindMin: Word;
function WindMax: Word;

//=============================================================================
// Standard methods
{$IFDEF fpas} procedure AssignCrt(var F: Text); {$ENDIF}
function KeyPressed: Boolean;
function ReadKey: Char;
procedure TextMode(
{$IFDEF fpas}Mode: SmallInt{$ELSE}SizeX, SizeY: SmallInt{$ENDIF});
procedure Window(X1, Y1, X2, Y2: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF});
procedure GotoXY(X, Y: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF});
function WhereX: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF};
function WhereY: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF};
procedure ClrScr;
procedure ClrEol;
procedure InsLine;
procedure DelLine;
procedure TextColor(Color: Byte);
procedure TextBackground(Color: Byte);
procedure LowVideo;
procedure HighVideo;
procedure NormVideo;
procedure Delay(MS: Word);
procedure Sound(Hz: Word);
{$IFDEF fpas} procedure NoSound; {$ENDIF}

//=============================================================================
// Additional types and methods

type
  pScreenBuffer = ^TScreenBuffer;
  TScreenBuffer = array of TCharInfo;

  TrEventType = (evtNone, evtKey, evtMouse, evtBuffer, evtFocus);
  TrConsoleEvent = record
    EventType: TrEventType;
    ControlKeyState: DWord;
    KeyRepeatCount: Word;
    KeyVirtualCode: Word;
    KeyChar: Char;
    MousePosition: TCoord;
    MouseButtonState: DWord;
    MouseEventFlags: DWord;
    BufferSize: TCoord;
    Focused: Boolean;
  end;

procedure SetTextAttr(Attr: Byte);
procedure Beep(dwFreq, dwDuration: DWORD);
procedure KeyCursorMode(aVisible: Boolean);
procedure MouseCursorMode(aVisible: Boolean);
procedure WaitForEvent(var aEvent: TrConsoleEvent);
function ReadLastKeyEvent: TKeyEventRecord;
function GetScreenSize: TCoord;
procedure PutSym(X, Y: SmallInt; Sym: Char);
procedure PutAttr(X, Y: SmallInt; Attr: Byte);
procedure PutSymAttr(X, Y: SmallInt; Sym: Char; Attr: Byte);
procedure FillRect(X1, Y1, X2, Y2: SmallInt; Sym: Char; Attr: Byte);

function GetTitle: String;
procedure SetTitle(aTitle: AnsiString);

// Screen buffer methods
procedure CopyScreenToBuffer;
procedure DrawScreenBuffer;
procedure BufferPutSym(X, Y: SmallInt; Sym: Char);
procedure BufferPutAttr(X, Y: SmallInt; Attr: Byte);
procedure BufferPutSymAttr(X, Y: SmallInt; Sym: Char; Attr: Byte);
procedure BufferFillRect(X1, Y1, X2, Y2: SmallInt; Sym: Char; Attr: Byte);

//=============================================================================
implementation

var
  ScreenSize, WinLT, WinRB: TCoord;
  ScreenBuffer: pScreenBuffer;
  hInp, hOut: THandle;
  _LastKeyEvent: TKeyEventRecord;

//=============================================================================
// Assigning a file variable for the console
{$IFDEF fpas}
procedure AssignCrt(var F: Text);
begin
  // not supported
end;
{$ENDIF}

//=============================================================================
// It is not recommended. Use WaitForEvent
function KeyPressed: Boolean;
var
  vInp: TInputRecord;
  dw: DWord;
begin
  Result := PeekConsoleInput(hInp, vInp, 1, dw) and
    (vInp.EventType = Key_Event) and
    (vInp.Event.KeyEvent.bKeyDown);
end;

//=============================================================================
// It is not recommended. Use WaitForEvent
function ReadKey: Char;
var
  vInp: TInputRecord;
  dw: DWord;
begin
  repeat until ReadConsoleInput(hInp, vInp, 1, dw) and
    (vInp.EventType = Key_Event) and
    (vInp.Event.KeyEvent.bKeyDown);
  _LastKeyEvent := vInp.Event.KeyEvent;
  Result := _LastKeyEvent.AsciiChar;
end;

// Read it when readkey returns #0
function ReadLastKeyEvent: TKeyEventRecord;
begin
  Result := _LastKeyEvent;
end;

//=============================================================================
// Changing the current coordinates
procedure GotoXY(X, Y: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF});
var
  Coord, WindowSize: TCoord;
begin
  WindowSize.X := WinRB.X - WinLT.X + 1;
  WindowSize.Y := WinRB.Y - WinLT.Y + 1;
  if (X >= 1) and (X <= WindowSize.X) and (Y >= 1) and (Y <= WindowSize.Y) then begin
    Coord.X := WinLT.X + X - 1;
    Coord.Y := WinLT.Y + Y - 1;
    SetConsoleCursorPosition(hOut, Coord);
  end;
end;

//=============================================================================
// Get the current coordinates

function WhereXY: TCoord;
var
  vInf: pConsoleScreenBufferInfo;
begin
  vInf := New(pConsoleScreenBufferInfo);
  try
    GetConsoleScreenBufferInfo(hOut, vInf^);
    Result := vInf^.dwCursorPosition;
  finally
    Dispose(vInf);
  end;
end;

function WhereX: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF};
begin
  Result := WhereXY.X + 1 - WinLT.X;
end;

function WhereY: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF};
begin
  Result := WhereXY.Y + 1 - WinLT.Y;
end;

//=============================================================================
// Clearing the screen
procedure ClrScr;
var
  c: TCoord;
  r: DWord;
  y: Byte;
begin
  for y := WinLT.Y to WinRB.Y do begin
    c.X := WinLT.X; c.Y := y;
    FillConsoleOutputCharacter(hOut, #32, WinRB.X - WinLT.X + 1, c, r);
    FillConsoleOutputAttribute(hOut, TextAttr, WinRB.X - WinLT.X + 1, c, r);
  end;
  SetConsoleCursorPosition(hOut, WinLT);
end;

//=============================================================================
// Cleaning up to the end of line
procedure ClrEol;
var
  c: TCoord;
  r: DWord;
begin
  c.X := WhereX - 1; c.Y := WhereY - 1;
  FillConsoleOutputCharacter(hOut, #32, WinRB.X - WhereX + 2, c, r);
  FillConsoleOutputAttribute(hOut, TextAttr, WinRB.X - WhereX + 2, c, r);
end;

//=============================================================================
// Insert and removal the line
procedure InsLine;
var
  //ClipRect: TSmallRect;
  SrcRect: TSmallRect;
  DestCoor: TCoord;
  CharInfo: TCharInfo;
  y: SmallInt;

begin
  y := WinLT.Y + WhereY;
  CharInfo.AsciiChar := #32;
  CharInfo.Attributes := _TextAttr;

  SrcRect.Left := WinLT.X;
  SrcRect.Top := y - 1;
  SrcRect.Right := WinRB.X;
  SrcRect.Bottom := WinRB.Y - 1;

  DestCoor.X := SrcRect.Left;
  DestCoor.Y := y;

  ScrollConsoleScreenBuffer(hOut, SrcRect, nil, DestCoor, CharInfo);
end;

procedure DelLine;
var
  SrcRect: TSmallRect;
  DestCoor: TCoord;
  CharInfo: TCharInfo;
  y: SmallInt;

begin
  y := WinLT.Y + WhereY;
  CharInfo.AsciiChar := #32;
  CharInfo.Attributes := _TextAttr;

  SrcRect.Left := WinLT.X;
  SrcRect.Top := y;
  SrcRect.Right := WinRB.X;
  SrcRect.Bottom := WinRB.Y;

  DestCoor.X := SrcRect.Left;
  DestCoor.Y := y - 1;

  ScrollConsoleScreenBuffer(hOut, SrcRect, nil, DestCoor, CharInfo);
end;

//=============================================================================
// Defining the working window
procedure Window(X1, Y1, X2, Y2: {$IFDEF fpas}Byte{$ELSE}SmallInt{$ENDIF});
begin
  if (X1 <= X2) and (Y1 <= Y2) and
     (X1 >= 1) and (X1 <= ScreenSize.X) and (Y1 >= 1) and (Y1 <= ScreenSize.Y) and
     (X2 >= 1) and (X2 <= ScreenSize.X) and (Y2 >= 1) and (Y2 <= ScreenSize.Y)
  then begin
    WinLT.X := X1 - 1; WinLT.Y := Y1 - 1;
    WinRB.X := X2 - 1; WinRB.Y := Y2 - 1;
  end;
  ClrScr;
end;

function WindMin: Word;
begin
  Result := ((WinLT.Y) Shl 8) + (WinLT.X);
end;

function WindMax: Word;
begin
  Result := ((WinRB.Y) Shl 8) + (WinRB.X);
end;

//=============================================================================
// Setting the attributes

function TextAttr: Byte;
begin
  Result := _TextAttr;
end;

procedure SetTextAttr(Attr: Byte);
begin
  SetConsoleTextAttribute(hOut, Attr);
  _TextAttr := Attr;
end;

procedure TextColor(Color: Byte);
begin
  SetTextAttr((TextAttr and $F0) or (Color and $0F));
end;

procedure TextBackground(Color: Byte);
begin
  SetTextAttr(((Color shl 4) and $F0) or (TextAttr and $0F));
end;

procedure LowVideo;
begin
  SetTextAttr(TextAttr and $F7);
end;

procedure HighVideo;
begin
  SetTextAttr(TextAttr or $08);
end;

procedure NormVideo;
begin
  SetTextAttr(TextAttr and $F7);
end;

//=============================================================================
// Delay and sound

procedure Delay(MS: Word);
begin
  Sleep(MS);
end;

procedure Sound(Hz: Word);
begin
  Beep(Hz, 100);
end;

{$IFDEF fpas}
procedure NoSound;
begin
  // not supported
end;
{$ENDIF}

//=============================================================================
// Setting the textual mode

procedure PrepareConsole;
begin
  SetLength(ScreenBuffer^, ScreenSize.X * ScreenSize.Y);
  WinLT.X := 0; WinLT.Y := 0;
  WinRB.X := ScreenSize.X - 1;
  WinRB.Y := ScreenSize.Y - 1;
  SetTextAttr(LightGray);
  ClrScr;
end;

procedure TextMode(
{$IFDEF fpas}Mode: SmallInt{$ELSE}SizeX, SizeY: SmallInt{$ENDIF});
{$IFNDEF fpas}
var
  c: TCoord;
{$ENDIF}
begin
{$IFDEF fpas}
  case Mode of
    BW40, CO40:       ScreenSize.X := 40;
    BW80, CO80, Mono: ScreenSize.X := 80
    else Exit
  end;
  ScreenSize.Y := 25;
  LastMode := Mode;
{$ELSE}
  ScreenSize.X := SizeX;
  ScreenSize.Y := SizeY;
  c.X := SizeX;
  c.Y := SizeY;
  SetConsoleScreenBufferSize(hOut, c);
{$ENDIF}
  PrepareConsole;
end;

//=============================================================================
// Sound signal
procedure Beep(dwFreq, dwDuration: DWORD);
begin
  Windows.Beep(dwFreq, dwDuration);
end;

//=============================================================================
// Setting keyboard cursor visibility
procedure KeyCursorMode(aVisible: Boolean);
var
  cInf: TConsoleCursorInfo;
begin
  GetConsoleCursorInfo(hOut, cInf);
  cInf.bVisible := BOOL(aVisible);
  SetConsoleCursorInfo(hOut, cInf);
end;

procedure MouseCursorMode(aVisible: Boolean);
begin
  // not needed on Win32 console
end;

//=============================================================================
// Waiting for the event

procedure WaitForEvent(var aEvent: TrConsoleEvent);

var
  Buff: TInputRecord;
  Records: DWORD;

begin
  aEvent.EventType := evtNone;
  repeat
    FlushConsoleInputBuffer(hInp);
    WaitForSingleObject(hInp, INFINITE);
    Records := 1;
    if ReadConsoleInput(hInp, Buff, 1, Records) then begin
      case Buff.EventType of
        KEY_EVENT:
          begin
            with Buff.Event.KeyEvent do
              if bKeyDown then begin
                aEvent.EventType := evtKey;
                aEvent.ControlKeyState := dwControlKeyState;
                aEvent.KeyRepeatCount := wRepeatCount;
                aEvent.KeyVirtualCode := wVirtualKeyCode;
                aEvent.KeyChar := AsciiChar;
              end;
          end;
        _MOUSE_EVENT:
          begin
            aEvent.EventType := evtMouse;
            with Buff.Event.MouseEvent do begin
              aEvent.ControlKeyState := dwControlKeyState;
              aEvent.MousePosition := dwMousePosition;
              aEvent.MouseButtonState := dwButtonState;
              aEvent.MouseEventFlags := dwEventFlags;
            end;
          end;
        WINDOW_BUFFER_SIZE_EVENT:
          begin
            aEvent.EventType := evtBuffer;
            with Buff.Event.WindowBufferSizeEvent do begin
              aEvent.EventType := evtBuffer;
              aEvent.BufferSize := dwSize;
            end;
          end;
        FOCUS_EVENT:
          begin
            aEvent.EventType := evtFocus;
            with Buff.Event.FocusEvent do begin
              aEvent.EventType := evtFocus;
              aEvent.Focused := bSetFocus;
            end;
          end;
      end;
    end;
  until aEvent.EventType <> evtNone;
end;

//=============================================================================
// Symbol and attributes output

procedure PutSym(X, Y: SmallInt; Sym: Char);
var
  cBuf: TCoord;
  rslt: DWORD;
begin
  cBuf.X := X - 1; cBuf.Y := Y - 1;
  WriteConsoleOutputCharacter(hOut, @Sym, 1, cBuf, rslt);
end;

procedure PutAttr(X, Y: SmallInt; Attr: Byte);
var
  AttrW: Word;
  cBuf: TCoord;
  rslt: DWORD;
begin
  cBuf.X := X - 1; cBuf.Y := Y - 1;
  AttrW := Attr;
  WriteConsoleOutputAttribute(hOut, @Attrw, 1, cBuf, rslt);
end;

procedure PutSymAttr(X, Y: SmallInt; Sym: Char; Attr: Byte);
var
  cBuf1,
  cBuf2: TCoord;
  Regn: TSmallRect;
  inf: TCharInfo;
begin
  cBuf1.X := 1; cBuf1.Y := 1;
  cBuf2.X := 0; cBuf2.Y := 0;
  Regn.Left := X - 1;
  Regn.Top := Y - 1;
  Regn.Right := X - 1;
  Regn.Bottom := Y - 1;
  inf.AsciiChar := Sym;
  inf.Attributes := Attr;
  WriteConsoleOutput(hOut, @inf, cBuf1, cBuf2, Regn);
end;

procedure FillRect(X1, Y1, X2, Y2: SmallInt; Sym: Char; Attr: Byte);
var
  c: TCoord;
  r: DWord;
  y: Byte;
begin
  for y := Y1 to Y2 do begin
    c.X := X1 - 1; c.Y := y - 1;
    FillConsoleOutputCharacter(hOut, Sym, X2 - X1 + 1, c, r);
    FillConsoleOutputAttribute(hOut, Attr, X2 - X1 + 1, c, r);
  end;
end;

//=============================================================================
// Screen to buffer exchange

procedure BufferExchange(aWrite: Boolean);
var
  cBuf1, cBuf2: TCoord;
  Regn: TSmallRect;
begin
  cBuf1.X := ScreenSize.X; cBuf1.Y := ScreenSize.Y;
  cBuf2.X := 0; cBuf2.Y := 0;
  Regn.Left := 0;
  Regn.Top := 0;
  Regn.Right := ScreenSize.X - 1;
  Regn.Bottom := ScreenSize.Y - 1;
  if aWrite
    then WriteConsoleOutput(hOut, ScreenBuffer, cBuf1, cBuf2, Regn)
    else ReadConsoleOutput(hOut, ScreenBuffer, cBuf1, cBuf2, Regn);
end;

procedure CopyScreenToBuffer;
begin
  BufferExchange(False);
end;

procedure DrawScreenBuffer;
begin
  BufferExchange(True);
end;

procedure BufferPutSym(X, Y: SmallInt; Sym: Char);
begin
  if (X >= 0) and (X <= ScreenSize.X) and (Y >= 0) and (Y <= ScreenSize.Y)
    then ScreenBuffer^[(X - 1)*(Y - 1)].AsciiChar := Sym;
end;

procedure BufferPutAttr(X, Y: SmallInt; Attr: Byte);
begin
  if (X >= 0) and (X <= ScreenSize.X) and (Y >= 0) and (Y <= ScreenSize.Y)
    then ScreenBuffer^[(X - 1)*(Y - 1)].Attributes := Attr;
end;

procedure BufferPutSymAttr(X, Y: SmallInt; Sym: Char; Attr: Byte);
begin
  BufferPutSym(X, Y, Sym);
  BufferPutAttr(X, Y, Attr);
end;

procedure BufferFillRect(X1, Y1, X2, Y2: SmallInt; Sym: Char; Attr: Byte);
var
  x, y: SmallInt;
begin
  for y := Y1 to Y2 do
    for x := X1 to X2 do
      BufferPutSymAttr(x, y, Sym, Attr);
end;

//=============================================================================

function GetScreenSize: TCoord;
begin
  Result := ScreenSize;
end;

function GetTitle: String;
var
  p: Pointer;
begin
  GetMem(p, 255);
  GetConsoleTitle(p, 255);
  Result := pChar(p);
  FreeMem(p);
end;

procedure SetTitle(aTitle: AnsiString);
begin
  SetConsoleTitle(pChar(aTitle));
end;

//=============================================================================
// Initialize and destroy

var
  vInf: pConsoleScreenBufferInfo;

initialization
  //SetConsoleOutputCP(GetACP);
//  SetFileApisToOEM;
  hInp := GetStdHandle(STD_INPUT_HANDLE);
  hOut := GetStdHandle(STD_OUTPUT_HANDLE);
  ScreenBuffer := New(pScreenBuffer);
{$IFDEF fpas}
  TextMode(CO80);
{$ELSE}
  vInf := New(pConsoleScreenBufferInfo);
  try
    GetConsoleScreenBufferInfo(hOut, vInf^);
    ScreenSize := vInf^.dwSize;
    PrepareConsole;
  finally
    Dispose(vInf);
  end;
{$ENDIF}

finalization
  CloseHandle(hOut);
  SetLength(ScreenBuffer^, 0);
  Dispose(ScreenBuffer);

//=============================================================================
end.

